// Copyright (c) 2004-2005 Lutz Latta
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

#define _USE_MATH_DEFINES

#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
using namespace std;

#include <math.h>
#include <expat/xmlparse.h>

#include "ParticleSystemConfig.h"


static float3 parseVector3(const char** attributes)
{
	float3 v;
	for (const char** a = attributes; *a; a += 2)
	{
		if (string(a[0]) == "x")
			v[0] = (float)atof(a[1]);
		else if (string(a[0]) == "y")
			v[1] = (float)atof(a[1]);
		else if (string(a[0]) == "z")
			v[2] = (float)atof(a[1]);
	}

	return v;
}

static float2 parseVector2(const char** attributes)
{
	float2 v;
	for (const char** a = attributes; *a; a += 2)
	{
		if (string(a[0]) == "x")
			v[0] = (float)atof(a[1]);
		else if (string(a[0]) == "y")
			v[1] = (float)atof(a[1]);
	}

	return v;
}

static float2 parseRange(const char** attributes)
{
	float2 v;
	for (const char** a = attributes; *a; a += 2)
	{
		if (string(a[0]) == "min")
			v[0] = (float)atof(a[1]);
		else if (string(a[0]) == "max")
			v[1] = (float)atof(a[1]);
	}

	return v;
}

static float4 parseColor(const char** attributes)
{
	float4 v;
	for (const char** a = attributes; *a; a += 2)
	{
		if (string(a[0]) == "red")
			v[0] = (float)atof(a[1]);
		else if (string(a[0]) == "green")
			v[1] = (float)atof(a[1]);
		else if (string(a[0]) == "blue")
			v[2] = (float)atof(a[1]);
		else if (string(a[0]) == "alpha")
			v[3] = (float)atof(a[1]);
	}

	return v;
}

static float parseScalar(const char** attributes, const string& name)
{
	for (const char** a = attributes; *a; a += 2)
	{
		if (name == a[0])
			return (float)atof(a[1]);
	}

	return 0.0f;
}

static string parseString(const char** attributes, const string& name)
{
	for (const char** a = attributes; *a; a += 2)
	{
		if (name == a[0])
			return a[1];
	}

	return "";
}


void ParticleSystemConfig::startElement(void *userData, const char *name, const char **attributes)
{
	ParticleSystemConfig* p = (ParticleSystemConfig*)userData;

	/*
	Parses something like this:
	<emitter>
		<position x="-1" y="3" z="0"/>
		<positionMaxOffset x="-1" y="3" z="0"/>
		<velocity min="0.15" max="0.17"/>
		<directionCenter x="0" y="-1" z="0"/>
		<directionDistibution maxAngle="15"/>
		<lifetime min="50" max="100"/>
		<attributes type="0"/>
		<numOfNewParticles min="50" max="150"/>
	</emitter>

	<replicateRotatedEmitter numOfEmitters="6" x="0" y="1" z="0" />

	<attributeType>
		<pointSize min="0.1" max="0.18" />
		<orientationStart min="45" max="90" />
		<orientationVelocity min="5" max="5" />
		<textureTileScale x="0.5" y="0.5" />
		<textureTileOffset x="0.5" y="0.5" />
		
		<!-- These need to be exactly four keyframes! -->
		<colorAnimationKeyframe time="0.0" red="-0.2" green="-0.02" blue="0.1" alpha="1.0" />
		<colorAnimationKeyframe time="0.1" red="1.0" green="0.1" blue="0.03" alpha="1.0" />
		<colorAnimationKeyframe time="0.6" red="0.5" green="0.1" blue="0.01" alpha="1.0" />
		<colorAnimationKeyframe time="1.0" red="0.0" green="0.0" blue="0.0" alpha="0.0" />
	</attributeType>

	<diffuseTexture name="fire_text_multi_alpha.dds" />

	<localForce x="0.0" y="0.0" z="0.0 />

	<flowFieldTexture name="flow_field.png" />
	<flowFieldScale x="4.0" y="1.0" z="4.0" />

	<sphereCollider x="1.0" y="1.0" z="0.0" />
	
	<viewer x="0.0" y="-2.0" z="-5.0" />
	<viewerRotation x="0.0" y="0.0" />
	
	<clearColor red="0.39" green="0.51" blue="0.59" alpha="1.0" />
	
	<!-- Possible types: "add", "alpha", "premultiplied-alpha" -->
	<blendFunction type="alpha" />
	
	<sorting enabled="true" />

	<particleSystemSize x="256" y="256" />
	
	<initialMenuKeys keys="sapl" />
	*/

	string n(name);
	if (p->ParserState == PS_MAIN)
	{
		if (n == "emitter")
		{
			p->ParserState = PS_EMITTER;
			p->Emitters.push_back(ParticleEmitterData());

			// PositionMaxOffset is optional.
			p->Emitters.back().PositionMaxOffset = float3(0, 0, 0);
		}
		else if (n == "replicateRotatedEmitter")
		{
			int numOfEmitters = (int)(parseScalar(attributes, "numOfEmitters") + 0.5f);
			float3 axis = parseVector3(attributes);

			// Replicate last emitter in circle around given axis.
			int originalEmitter = p->Emitters.size() - 1;
			p->Emitters.resize(originalEmitter + numOfEmitters, p->Emitters[originalEmitter]);
			for (int i = 1; i < numOfEmitters; i++)
			{
				glh::quaternionf q(axis, (float)M_PI * 2.0f / numOfEmitters * i);

				ParticleEmitterData& e = p->Emitters[originalEmitter + i];
				q.mult_vec(e.Position);
				e.PreviousPosition = e.Position;
				q.mult_vec(e.InitialDirectionCenter);
			}
		}
		else if (n == "attributeType")
		{
			p->ParserState = PS_ATTRIBUTES;
			p->ParticleAttributes.push_back(ParticleAttributesType());
		}
		else if (n == "diffuseTexture")
		{
			p->DiffuseTextureName = parseString(attributes, "name");
		}
		else if (n == "localForce")
		{
			p->LocalForcePositions.push_back(parseVector3(attributes));
		}
		else if (n == "flowFieldTexture")
		{
			p->FlowFieldTextureName = parseString(attributes, "name");
		}
		else if (n == "flowFieldScale")
		{
			p->FlowFieldScale = parseVector3(attributes);
		}
		else if (n == "sphereCollider")
		{
			p->SphereColliderPosition = parseVector3(attributes);
		}
		else if (n == "viewer")
		{
			p->ViewerPosition = parseVector3(attributes);
		}
		else if (n == "viewerRotation")
		{
			p->ViewerRotation = parseVector2(attributes);
		}
		else if (n == "clearColor")
		{
			p->ClearColor = parseColor(attributes);
		}
		else if (n == "blendFunction")
		{
			string s = parseString(attributes, "type");
			if (s == "add")
				p->BlendFunction = ParticleSystem::BF_ADD;
			else if (s == "alpha")
				p->BlendFunction = ParticleSystem::BF_ALPHA;
			else if (s == "premultiplied-alpha")
				p->BlendFunction = ParticleSystem::BF_PREMULTIPLIED_ALPHA;
		}
		else if (n == "sorting")
		{
			p->DoSorting = parseString(attributes, "enabled") == "true";
		}
		else if (n == "particleSystemSize")
		{
			float2 v = parseVector2(attributes);
			p->Width = int(v[0] + 0.5f);
			p->Height = int(v[1] + 0.5f);
		}
		else if (n == "initialMenuKeys")
		{
			string keys = parseString(attributes, "keys");
			for (int i = 0; i < 256; i++)
				p->MenuKeys[i] = false;
			for (string::iterator c = keys.begin(); c != keys.end(); c++)
				p->MenuKeys[*c] = true;
		}
	}
	else if (p->ParserState == PS_EMITTER)
	{
		ParticleEmitterData& e = p->Emitters.back();
		if (n == "position")
		{
			e.Position = parseVector3(attributes);
			e.PreviousPosition = e.Position;
		}
		if (n == "positionMaxOffset")
			e.PositionMaxOffset = parseVector3(attributes);
		else if (n == "velocity")
			e.InitialVelocityMinMax = parseRange(attributes);
		else if (n == "directionCenter")
		{
			e.InitialDirectionCenter = parseVector3(attributes);
			e.InitialDirectionCenter.normalize();
		}
		else if (n == "directionDistibution")
			e.InitialDirectionMaxAngle = parseScalar(attributes, "maxAngle") * (float)M_PI / 180.0f;
		else if (n == "lifetime")
			e.LifeTimeMinMax = parseRange(attributes);
		else if (n == "attributes")
			e.AttributeType = (int)(parseScalar(attributes, "type") + 0.5f);
		else if (n == "numOfNewParticles")
			e.NumOfNewParticlesMinMax = parseRange(attributes);
	}
	else if (p->ParserState == PS_ATTRIBUTES)
	{
		ParticleAttributesType& a = p->ParticleAttributes.back();
		if (n == "pointSize")
		{
			float2 v = parseRange(attributes);
			a.PointSizeRange = float2(v[0], v[1] - v[0]);
		}
		else if (n == "orientationStart")
		{
			float2 v = parseRange(attributes);
			a.OrientationStartRange_VelocityRange[0] = v[0] * (float)M_PI / 180.0f;
			a.OrientationStartRange_VelocityRange[1] = (v[1] - v[0]) * (float)M_PI / 180.0f;
		}
		else if (n == "orientationVelocity")
		{
			float2 v = parseRange(attributes);
			a.OrientationStartRange_VelocityRange[2] = v[0] * (float)M_PI / 180.0f;
			a.OrientationStartRange_VelocityRange[3] = (v[1] - v[0]) * (float)M_PI / 180.0f;
		}
		else if (n == "textureTileScale")
		{
			float2 v = parseVector2(attributes);
			a.TileScale_TileOffset[0] = v[0];
			a.TileScale_TileOffset[1] = v[1];
		}
		else if (n == "textureTileOffset")
		{
			float2 v = parseVector2(attributes);
			a.TileScale_TileOffset[2] = v[0];
			a.TileScale_TileOffset[3] = v[1];
		}
		else if (n == "colorAnimationKeyframe")
		{
			p->CurrentKeyframeTimes[p->CurrentKeyframeIndex] = parseScalar(attributes, "time");
			p->CurrentKeyframes[p->CurrentKeyframeIndex] = parseColor(attributes);

			p->CurrentKeyframeIndex++;

			if (p->CurrentKeyframeIndex == 4)
			{
				a.ColorAnimationSegments = p->CurrentKeyframeTimes;
				for (int j = 0; j < 3; j++)
				{
					// Linear function segments y = m * x + b,
					// encoded with m in even and b in odd indices.

					// m = (y1 - y0) / (x1 - x0)
					a.ColorAnimationFunctions[j * 2] 
						= (p->CurrentKeyframes[j + 1] - p->CurrentKeyframes[j])
						/ (p->CurrentKeyframeTimes[j + 1] - p->CurrentKeyframeTimes[j]);
					// b = y0 - m * x0
					a.ColorAnimationFunctions[j * 2 + 1] 
						= p->CurrentKeyframes[j] - a.ColorAnimationFunctions[j * 2] * p->CurrentKeyframeTimes[j];
				}

				p->CurrentKeyframeIndex = 0;
			}
		}
	}

}

void ParticleSystemConfig::endElement(void *userData, const char *name)
{
	ParticleSystemConfig* p = (ParticleSystemConfig*)userData;

	string n(name);
	if (p->ParserState == PS_EMITTER)
	{
		if (n == "emitter")
			p->ParserState = PS_MAIN;
	}
	else if (p->ParserState == PS_ATTRIBUTES)
	{
		if (n == "attributeType")
			p->ParserState = PS_MAIN;
	}

}

void ParticleSystemConfig::ReadConfig(const string& fileName)
{
	data_path path;
	path.path.push_back(".");
	path.path.push_back("./shaders");
	path.path.push_back("../media/shaders");
	path.path.push_back("../../media/shaders");

	std::string pathName = path.get_file(fileName);
	if (pathName == "") {
		cerr << "Unable to load '" + fileName + "', exiting..." << endl;
		exit(-1);
	}

	ifstream f(pathName.c_str(), ios_base::binary);

	vector<char> buf;
	f.seekg(0, ios_base::end);
	buf.resize(f.tellg());
	f.seekg(0);

	f.read(&buf.front(), buf.size());
	f.close();

	XML_Parser parser = XML_ParserCreate(NULL);

	XML_SetUserData(parser, this);

	XML_SetElementHandler(parser, startElement, endElement);
	if (!XML_Parse(parser, &buf.front(), buf.size(), true))
	{
		cerr << "XML Parser Error: " << XML_ErrorString(XML_GetErrorCode(parser)) << " in file '" << fileName << "' at line " << XML_GetCurrentLineNumber(parser) << endl;
		exit(1);
	}
	XML_ParserFree(parser);
}
