// Copyright (c) 2004-2005 Lutz Latta
//
// Permission is hereby granted, free of charge, to any person obtaining
// a copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.

#include <math.h>
#include <string>
#include <iostream>
#include <fstream>
#include <glh/glh_extensions.h>

using namespace std;

#include "SimpleMesh.h"

SimpleMesh::SimpleMesh()
{ 
	glGenBuffersARB(1, &VertexBuffer);
	glGenBuffersARB(1, &IndexBuffer);
}

SimpleMesh::~SimpleMesh()
{
	glDeleteBuffersARB(1, &VertexBuffer);
	glDeleteBuffersARB(1, &IndexBuffer);
}

void SimpleMesh::Load(string const& fileName)
{
	if (fileName.substr(fileName.size() - 8) == ".meshpos"
		|| fileName.substr(fileName.size() - 12) == ".meshposnorm")
	{
		bool hasNormals = fileName.substr(fileName.size() - 12) == ".meshposnorm";

		ifstream inFile(fileName.c_str(), ios::in);

		char buf[256] = { 0 };
		inFile.getline(buf, sizeof(buf), ' ');
		if (string(buf) != "VERTEX:")
		{
			cerr << "Invalid or not-found mesh file: " << fileName << endl;
			exit(1);
		}

		int num;
		inFile >> num;
		Vertices.resize(num);
		for (int i = 0; i < num; i++)
		{
			inFile >> Vertices[i].Position[0] >> Vertices[i].Position[1] >> Vertices[i].Position[2];
			if (hasNormals)
				inFile >> Vertices[i].Normal[0] >> Vertices[i].Normal[1] >> Vertices[i].Normal[2];
		}

		inFile.getline(buf, sizeof(buf));
		inFile.getline(buf, sizeof(buf), ' ');
		if (string(buf) != "INDEX:")
		{
			cerr << "Invalid mesh file: " << fileName << endl;
			exit(1);
		}

		inFile >> num;
		num *= 3;
		Indices.resize(num);
		for (int i = 0; i < num; i++)
			inFile >> Indices[i];

		inFile.close();

		if (!hasNormals)
			GenerateNormals();
	}
	else
	{
		ifstream binaryInFile(fileName.c_str(), ios::in | ios::binary);

		char buf[10] = { 0 };
		binaryInFile.read(buf, sizeof(buf));
		if (strncmp(buf, "simplemesh", sizeof(buf)))
		{
			cerr << "Invalid or not-found mesh file: " << fileName << endl;
			exit(1);
		}

		int num;
		binaryInFile.read((char*)&num, sizeof(num));
		Vertices.resize(num);
		for (vector<Vertex>::iterator i = Vertices.begin(); i != Vertices.end(); i++)
		{
			binaryInFile.read((char*)&i->Position, sizeof(float3));
			binaryInFile.read((char*)&i->Normal, sizeof(float3));
		}

		binaryInFile.read((char*)&num, sizeof(num));
		Indices.resize(num);
		binaryInFile.read((char*)&Indices.front(), sizeof(Indices[0]) * num);

		binaryInFile.close();
	}

	UploadBuffers();
}

void SimpleMesh::Save(string const& fileName)
{
	ofstream binaryOutFile(fileName.c_str(), ios::out | ios::binary);
	binaryOutFile.write((char*)"simplemesh", sizeof("simplemesh") - 1);
	int num = Vertices.size();
	binaryOutFile.write((char*)&num, sizeof(num));
	for (vector<Vertex>::iterator i = Vertices.begin(); i != Vertices.end(); i++)
	{
		binaryOutFile.write((char*)&i->Position, sizeof(float3));
		binaryOutFile.write((char*)&i->Normal, sizeof(float3));
	}
	num = Indices.size();
	binaryOutFile.write((char*)&num, sizeof(num));
	binaryOutFile.write((char*)&Indices.front(), sizeof(Indices[0]) * num);
	binaryOutFile.close();
}

void SimpleMesh::TransformVertices(float4x4 const& transformation)
{
	float4x4 inverseTranspose = transformation.inverse().transpose();

	for (vector<Vertex>::iterator i = Vertices.begin(); i != Vertices.end(); i++)
	{
		transformation.mult_matrix_vec(i->Position);
		inverseTranspose.mult_matrix_dir(i->Normal);
		i->Normal.normalize();
	}

	UploadBuffers();
}

// Normalize bounding box of mesh to [-1,1]^3
void SimpleMesh::NormalizeSizeAndCenter(void)
{
	if (Vertices.empty())
		return;

	float3 minPos = Vertices[0].Position;
	float3 maxPos = Vertices[0].Position;

	for (vector<Vertex>::iterator i = Vertices.begin(); i != Vertices.end(); i++)
	{
		for (int c = 0; c < 3; c++)
		{
			if (minPos[c] > i->Position[c])
				minPos[c] = i->Position[c];
			if (maxPos[c] < i->Position[c])
				maxPos[c] = i->Position[c];
		}
	}

	float3 extent = maxPos - minPos;
	float maxExtent = max(extent[0], max(extent[1], extent[2]));

	float3 shift = -(maxPos + minPos) / 2.0f;

	float4x4 m;
	m.set_translate(shift);

	float4x4 t;
	t.set_scale(2.0f / maxExtent);

	m = m.mult_left(t);

	TransformVertices(m);
}

// Generate mesh normals.
void SimpleMesh::GenerateNormals(void)
{
	for (vector<Vertex>::iterator i = Vertices.begin(); i != Vertices.end(); i++)
		for (int c = 0; c < 3; c++)
			i->Normal[c] = 0.0f;

	float3 normal;
	for (int i = 0; i < (int)Indices.size(); i += 3)
	{
		normal = float3(Vertices[Indices[i + 1]].Position - Vertices[Indices[i]].Position).
			cross(Vertices[Indices[i + 2]].Position - Vertices[Indices[i]].Position);
		normal.normalize();

		for (int c = 0; c < 3; c++)
			Vertices[Indices[i + c]].Normal = Vertices[Indices[i + c]].Normal + normal;
	}

	for (vector<Vertex>::iterator i = Vertices.begin(); i != Vertices.end(); i++)
		for (int c = 0; c < 3; c++)
			i->Normal.normalize();

	UploadBuffers();
}

void SimpleMesh::Render()
{
	glPushClientAttrib(GL_CLIENT_VERTEX_ARRAY_BIT);


	glBindBufferARB(GL_ARRAY_BUFFER_ARB, VertexBuffer);

	Vertex dummyVertex;

	glNormalPointer(GL_FLOAT, sizeof(Vertex), (GLvoid*)((char*)&dummyVertex.Normal - (char*)&dummyVertex));
	glEnableClientState(GL_NORMAL_ARRAY);

	glVertexPointer(3, GL_FLOAT, sizeof(Vertex), (GLvoid*)((char*)&dummyVertex.Position - (char*)&dummyVertex));
	glEnableClientState(GL_VERTEX_ARRAY);

	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);


	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, IndexBuffer);

	glDrawElements(GL_TRIANGLES, Indices.size(), GL_UNSIGNED_SHORT, 0);

	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);


	glPopClientAttrib();
}

// Create mesh from height field.
void SimpleMesh::CreateFromHeightField(const array2<unsigned char>& heightField)
{
	Vertices.clear();
	Vertices.resize(heightField.get_width() * heightField.get_height());
	Indices.clear();
	Indices.resize((heightField.get_width() - 1) * (heightField.get_height() - 1) * 6);

	vector<Vertex>::iterator v = Vertices.begin();
	for (int y = 0; y < heightField.get_height(); y++)
		for (int x = 0; x < heightField.get_width(); x++, v++)
			v->Position = float3((float)x, (float)heightField(x, y), (float)y);

	vector<unsigned short>::iterator d = Indices.begin();
	int width = heightField.get_width();
	for (int y = 0; y < heightField.get_height() - 1; y++)
	{
		for (int x = 0; x < width - 1; x++)
		{
			*(d++) = y * width + x;
			*(d++) = y * width + x + 1;
			*(d++) = (y + 1) * width + x + 1;

			*(d++) = (y + 1) * width + x + 1;
			*(d++) = (y + 1) * width + x;
			*(d++) = y * width + x;
		}
	}

	GenerateNormals();
}

void SimpleMesh::UploadBuffers()
{
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, VertexBuffer);
	glBufferDataARB(GL_ARRAY_BUFFER_ARB, Vertices.size() * sizeof(Vertex), &Vertices.front(), GL_STATIC_DRAW_ARB);
	glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);

	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, IndexBuffer);
	glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB, Indices.size() * sizeof(unsigned short), &Indices.front(), GL_STATIC_DRAW_ARB);
	glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, 0);
}
